;;; bijin-tokei.el --- bijin-tokei on Emacs
;;;
;;; Version: 20150302
;;;
;;; Code:

(require 'deferred)

(defvar bijin-tokei-last-display-time ""
  "美人時計の最終アップデート時間")

(defvar bijin-tokei-timer nil
  "美人時計の定期実行用のタイマー")

(defvar bijin-tokei-url-prefix
  "http://www.bijint.com/assets/pict/"
  "美人時計のURLのプレフィクス")

(defcustom bijin-tokei-genre
  "jp"
  "美人時計のジャンル"
  :group 'bijin-tokei)

(defvar bijin-tokei-url-mid
  "/pc/"
  "美人時計のURLの中間部品")

(defvar bijin-tokei-url-postfix ".jpg"
  "美人時計のURLのポストフィックス(拡張子)")

(defcustom bijin-tokei-genre-list (list
                                "jp"
                                "tokyo"
                                "osaka"
                                "nagasaki"
                                "kanazawa"
                                "hiroshima"
                                "kagawa"
                                "tochigi"
                                "fukui"
                                "akita"
                                "gunma"
                                "chiba"
                                "hokkaido"
                                "kobe"
                                "niigata"
                                "kanagawa"
                                "megane"
                                "tv-asahi"
                                "cc"
                                "binan")
  "美人時計のジャンルのリスト"
  :group 'bijin-tokei)

(defcustom bijin-tokei-shuffle-genre nil
  "ジャンルをシャッフルする"
  :group 'bijin-tokei)

(defvar bijin-tokei-buffername "*bijin-tokei*"
  "美人時計バッファ名")

(defvar bijin-tokei-mode-line-format
  '("    "
    mode-line-frame-identification
    mode-line-buffer-identification))

(defvar bijin-tokei-url-access-retry      0)
(defvar bijin-tokei-url-access-retry-max 99)
(defvar bijin-tokei-debug nil)

(defun bijin-tokei-start ()
  "美人時計の開始"
  (interactive)
  (if (eq (window-system) nil)
      (message "Sorry it can't display in no-window-system")
    (progn
      (setq bijin-tokei-url-access-retry 0)
      (setq bijin-tokei-last-display-time "")
      (setq bijin-tokei-timer (run-with-timer t 1 'bijin-tokei))
      (let ((selected))
        (setq selected (selected-window))
        (switch-to-buffer-other-window bijin-tokei-buffername)
        (select-window selected)))))

(defun bijin-tokei-stop ()
  "美人時計の停止"
  (interactive)
  (cancel-timer bijin-tokei-timer)
  (kill-buffer bijin-tokei-buffername))

(defun bijin-tokei ()
  "更新の必要のあるものだけ"
  (if (not
      (equal
       bijin-tokei-last-display-time
       (format-time-string "%H%M" (current-time))))
      (bijin-tokei-update-buffer-async)
    ))

(defun bijin-tokei-update-buffer-async ()
  "美人時計 表示バッファを更新する"
  (let* ((url (concat bijin-tokei-url-prefix
                      (if bijin-tokei-shuffle-genre
                          (nth (random (length bijin-tokei-genre-list)) bijin-tokei-genre-list)
                        bijin-tokei-genre)
                      bijin-tokei-url-mid
                      (format-time-string "%H%M" (current-time))
                      bijin-tokei-url-postfix)))
    (if bijin-tokei-debug
        (progn (setq url-show-status t)
               (message (format "%s" url)))
      (setq url-show-status nil))
    (deferred:$
      (deferred:url-retrieve url)
      (deferred:nextc it
        (lambda (buf)
          (let ((bijin-jpg-data (with-current-buffer buf (buffer-string))))
            (setq bijin-jpg-data (substring bijin-jpg-data (+ (string-match "\n\n" bijin-jpg-data) 2)))
            (with-current-buffer (get-buffer-create bijin-tokei-buffername)
              (setq cursor-type nil)
              (read-only-mode 0)
              (buffer-disable-undo)
              (erase-buffer)
              (goto-char (point-min))
              (let* ((bjsize (window-inside-pixel-edges (car (get-buffer-window-list bijin-tokei-buffername))))
                     (width  (+ (- (nth 2 bjsize) (nth 0 bjsize)) 1))
                     (height (+ (- (nth 3 bjsize) (nth 1 bjsize)) 1)))
                (if (and (> width 0) (> height 0))
                    (insert-image (bjimg--fit-to-size (create-image bijin-jpg-data 'jpeg :data) width height 3))))
              (setq mode-line-format bijin-tokei-mode-line-format)
              (read-only-mode 1))
            (setq bijin-tokei-url-access-retry 0)
            (setq bijin-tokei-last-display-time
                  (format-time-string "%H%M" (current-time))))))
      (deferred:error it ;
        (lambda (err)
          (progn
            (if (<= bijin-tokei-url-access-retry bijin-tokei-url-access-retry-max)
                (progn
                  (setq bijin-tokei-url-access-retry (+ bijin-tokei-url-access-retry 1))
                  (if bijin-tokei-debug (message (format "retrying %s" bijin-tokei-url-access-retry))))
              (progn
                (if bijin-tokei-debug (message (format "retry-max %s" bijin-tokei-url-access-retry)))
                (setq bijin-tokei-url-access-retry 0)
                (setq bijin-tokei-last-display-time
                      (format-time-string "%H%M" (current-time))))))
          ))
      )
    ))

(defcustom bijin-tokei-image-convert-command "convert"
  "ImageMagick convert command"
  :group 'bijin-tokei
  :type 'file)

(defun bjimg--call-convert (image &rest args)
  (let ((spec (cdr image)))
    (with-temp-buffer
      (set-buffer-multibyte nil)
      (let ((coding-system-for-read 'binary)
            (coding-system-for-write 'binary))
        (cond
         ((plist-get spec :data)
          (insert (plist-get spec :data))
          ;; stdin to current-buffer
          (unless (eq (apply 'call-process-region
                             (point-min) (point-max) bijin-tokei-image-convert-command
                             t (current-buffer) nil
                             `(,@args "-" "-")) 0)
            (error "Cannot convert image")))
         ((plist-get spec :file)
          ;; stdin to current-buffer
          (unless (eq (apply 'call-process
                             bijin-tokei-image-convert-command
                             (plist-get spec :file) t nil
                             `(,@args "-" "-")) 0)
            (error "Cannot convert image")))
         (t
          (error "Not a supported image")))
        (let ((img (create-image (buffer-string) nil t)))
          (plist-put (cdr img) 'bijin-tokei-original-image
                     (or (plist-get (cdr image) 'bijin-tokei-original-image)
                         (copy-sequence image)))
          img)))))

(defun bjimg--zoom (image magnification)
  (let* ((pixels (image-size image t))
         (new (bjimg--call-convert
               image
               "-resize"
               (format "%sx%s"
                       (truncate (* (car pixels) magnification))
                       (truncate (* (cdr pixels) magnification))))))
    ;; clone source image properties
    (when (plist-get (cdr image) :margin)
      (plist-put (cdr new) :margin
                 (plist-get (cdr image) :margin)))
    (when (plist-get (cdr image) :relief)
      (plist-put (cdr new) :relief
                 (plist-get (cdr image) :relief)))
    new))

(defun bjimg--fit-to-size (image width height &optional max)
  "Resize IMAGE with preserving magnification."
  (let* ((pixels (image-size image t))
         (margin (or (plist-get (cdr image) :margin) 0))
         (relief (or (plist-get (cdr image) :relief) 0))
         (mr (+ (* 2 margin) (* 2 relief)))
         (w (+ (car pixels) mr))
         (h (+ (cdr pixels) mr))
         (wr (/ width (ftruncate w)))
         (hr (/ height (ftruncate h)))
         (magnification (min wr hr max)))
    (bjimg--zoom image magnification)))


(provide 'bijin-tokei)
